import numpy as np
from ast import literal_eval


# liest den event tree aus, man muss das voll schlüsselwort angeben
def getData(eventTree, keyword: str):
    try:
        return eventTree.arrays(keyword, library='np')[keyword]
    except:
        return KeyError


# generiert für jeden cluster eine event nummer
def genEventNumbers(clusters):
    eventNumbers = []
    for i in range(len(clusters)):
        eventNumbers.append(np.array([i]*len(clusters[i])))
    return flatten(eventNumbers)


# organisiert mc und digit daten aller events um, sodass sie zu clustern passen
def getEventData(relations, *args):
    returnList = []
    for i in range(len(args)):
        stuffList = []
        for item in relations:
            stuffList.append([0] * len(item))
        returnList.append(stuffList)
    for i, references in enumerate(relations):
        for k, index in enumerate(references):
            for j in range(len(args)):
                returnList[j][i][k] = args[j][index]
    if len(returnList) == 1:
        return returnList[0]
    else:
        return returnList


# reduziert die dimension von arrays/listen um eins
def flatten(metric):
    length = 0
    for element in metric:
        length += len(element)
    flatMetric = [0] * length
    start = 0
    for i in range(len(metric)):
        stop = len(metric[i])
        flatMetric[start:start+stop] = metric[i]
        start += stop
    return np.array(flatMetric, dtype=object)


# this function flattens python lists, since they can be ragged and numpy/pytorch have troubles handling them
# List: list => the ragged list to be flattened
# list -> it returns a flat list
def flattenList(List: list) -> list:
    returnList = []
    for item in List:
        if type(item) == list:
            elements = flatten(item)
            for element in elements:
                returnList.append(element)
        else:
            returnList.append(item)
    return returnList


# generiert alle pixel matrizen aller events
def getClustersFlattened(uCellIDs, vCellIDs, cellCharges, clusterDigits, matrixSize=(9,9)):
    length = 0
    start = 0
    for item in cellCharges:
        length += len(item)
    events = [0] * length
    plotRange = int(np.round(matrixSize[0]/2)), int(np.round(matrixSize[1]/2))
    for event in range(len(cellCharges)):
        adcValues = []
        digitsU = uCellIDs[event]
        digitsV = vCellIDs[event]
        digitsCharge = cellCharges[event]
        digitIndices = clusterDigits[event]
        for indices in digitIndices:
            cacheImg = np.zeros(matrixSize)
            maxChargeIndex = digitsCharge[indices].argmax()
            uMax = digitsU[indices[maxChargeIndex]]
            vMax = digitsV[indices[maxChargeIndex]]
            for index in indices:
                uPos = digitsU[index]
                vPos = digitsV[index]
                uu = int(uPos) - int(uMax) + plotRange[0]
                vv = int(vPos) - int(vMax) + plotRange[1]
                if uu >= 0 and uu < matrixSize[0] and vv >= 0 and vv < matrixSize[1]:
                    cacheImg[uu,vv] = digitsCharge[index]
            adcValues.append(cacheImg)
        stop = len(adcValues)
        events[start:start+stop] = adcValues
        start += stop
    return np.array(events, dtype=object)


# rotiert und verschiebt eine Koordinate
def rotShiftVector(vector, angle, shift=[0,0,0], scale=1):
    theta = np.deg2rad(angle)
    rotMatrix = np.array([[np.cos(theta),-np.sin(theta),0],[np.sin(theta),np.cos(theta),0],[0,0,1]])
    scaleMatrix = np.array([[1,0,0],[0,1,0],[0,0,scale]])
    return rotMatrix.dot(scaleMatrix.dot(vector)) + shift


# berechnet die koordinaten aller Events
def getCartesianFlattened(uPositions, vPositions, sensorIDs, transformations: dict):
    length = 0
    start = 0
    for item in sensorIDs:
        length += len(item)
    xArr, yArr, zArr = [0] * length, [0] * length, [0] * length
    for event in range(len(sensorIDs)):
        xyz = []
        uPos = uPositions[event]
        vPos = vPositions[event]
        sensors = sensorIDs[event]
        points = np.vstack((uPos, np.zeros(len(uPos)), vPos)).T
        for point, id in zip(points, sensors):
            shift, angle = transformations[str(id)]
            shifted = rotShiftVector(point, angle, shift)
            xyz.append(shifted)
        if len(xyz) > 0:
            stop = len(xyz)
            xArr[start:start+stop] = np.array(xyz)[:,0]
            yArr[start:start+stop] = np.array(xyz)[:,1]
            zArr[start:start+stop] = np.array(xyz)[:,2]
        start += stop
    return np.array(xArr, dtype=object), np.array(yArr, dtype=object), np.array(zArr, dtype=object)


# bestimmt die layer und ladder nummer eines events
def getLayers(sensorIDs, layersLadders: dict):
    layers, ladders = [], []
    for id in sensorIDs:
        layer, ladder = layersLadders[str(id)]
        layers.append(layer)
        ladders.append(ladder)
    return np.array(layers), np.array(ladders)


# findet die fehlenden event referenzen in mc daten, setzt sie gleich -1
def findMissing(lst: list, length: int) -> list:
    return sorted(set(range(0, length)) - set(lst))


# füllt die mc-cluster beziehungs arrays mit fehlenden werten auf
def fillMCList(fromClusters, toClusters, length):
    missingIndex = findMissing(fromClusters, length)
    testList = [-1] * length
    fillIndex = 0
    for i in range(len(testList)):
        if i in missingIndex:
            testList[i] = -1
        else:
            try:
                testList[i] = int(toClusters[fillIndex])
            except TypeError:
                testList[i] = int(toClusters[fillIndex][0])
            fillIndex += 1
    return testList


# organisiert mc daten eines events um
def getMCData(toClusters, pdgs, xMom, yMom, zMom):
    pxList, pyList, pzList = [], [], []
    pdgList = []
    for references in toClusters:
        if references == -1:
            pxList.append(0)
            pyList.append(0)
            pzList.append(0)
            pdgList.append(0)
        else:
            pxList.append(xMom[references])
            pyList.append(yMom[references])
            pzList.append(zMom[references])
            pdgList.append(pdgs[references])
    return np.array(pdgList,dtype=list), np.array(pxList,dtype=list), np.array(pyList,dtype=list), np.array(pzList,dtype=list)


# this function returns a variable as the type it is meant to be used
# I am using it to assess variables from config files
# input: str => all kinds of data types as strings
# input -> returns the input as the correct datatype
def getType(input: str):
    # first this function tries to guess a data type, it will fail in cases of strings
    try:
        input = literal_eval(input)
        # here tuples will be converted into lists, simply because literal_eval seems to prefer tuples over lists, I don't
        if type(input) == tuple:
            return list(input)
        else:
            return input
    # here everything that has been deemed to be a string will be parsed, either they are strings or strings in lists
    except (ValueError, SyntaxError):
        input = input.replace(' ','').replace(';', ',') # getting rid of semi colons and spaces
        # this is is the user writes a list with brackets, a bit more of work is needed to turn it into potentially nested lists
        if '[' in input:
            input = input.replace('[','')
            input = input.split('],')
            output = []
            for line in input:
                line = line.replace(']','')
                output.append(line.split(','))
            return output
        # this just splits a list of strings into a list full of strings
        elif ',' in input:
            return input.split(',')
        # just returns the string as is
        else:
            return input