from baseClass import parametersClass
from functions import flatten, flattenList, getType
import os
from glob import glob


# this class handles a SINGLE parameter of the network, it takes care of options and types, so that nothing gets assigned incorretly
# check: tests of a given input makes sense for this parameter
#  help: returns a few usefull information
class settingBase():
    #          name: str => the name of the parameter
    #       default: whatever => the default valued held by this parameter, it falls back onto this if user input in incorrect
    #       tooltip: str => a description of what this parameter is doing
    #         types: list => this contains the possible datatypes this parameter can be, it is inferred from default if nothing is given
    #       options: list => a list of all possible options a parameter can be set to
    # convertToList: bool => whether or not to convert given input into a list of that input, this is imported for some options with multiply settings
    def __init__(self, name: str, default, tooltip: str = '', types: list = [None], options: list = [None], convertToList: bool = False):
        self.name = name
        self.tooltip = tooltip
        self.value = default
        self.default = default
        self.convertToList = convertToList
        if types == [None]:
            self.types = [type(default)]
        else:
            self.types = flattenList([types])
        self.options = options
        if type(default) == bool and options == [None]:
            self.options = [True, False]

    # this checks if an input makes sense for this parameter, no assigning is happening here
    # value: whatever => the value one wants to assign
    # (bool, bool) -> returns tuple of bools, first is if the type is correct and second if all the options make sense
    def checks(self, value) -> (bool, bool):
        typeBool, optionBool = False, False
        if type(value) in self.types:
            typeBool = True
        if isinstance(value, list) and not self.options == [None]:
            value = flatten(value)
            checks = [False]*len(value)
            for i,item in enumerate(value):
                checks[i] = item in self.options
            optionBool = all(checks)
        elif value in self.options or self.options == [None]:
            optionBool = True
        return typeBool, optionBool

    # this just return some information about this setting/parameter
    def help(self) -> tuple:
        return self.name, self.types, self.options, self.tooltip

    def __repr__(self) -> str:
        return str(self.value)

    def __len__(self) -> int:
        return len(self.value)


# this class finally contains parameters, but only of single section, it handels the creation and setting of the actual parameters
# the networks/programms parameter are variables/parameters of this class, which can be called like normal values/variables
# it also stores all incorrectly user parameters
# it handles the printing and writing of its own section parameters
class sectionClass(parametersClass):
    # section: str => the name of the section of parameters
    def __init__(self, section: str):
        self.__sectionName__ = section
        self.__maxLen__ = 0
        self.__wrong__ = {}
        self.__numAttr__ = 0

    # this sets/assigns parameters to this class, calls functions of the settingBase class inorder to determine if a user input can be assigned
    #   key: str => the name of the parameter
    # value: whatever => the value one wants to store
    def __setitem__(self, key: str, value):
        if key not in self.list():
            self.__numAttr__ += 1
            if type(value) == tuple:
                self.setValue(key, settingBase(key, *value))
            else:
                self.setValue(key, settingBase(key, value))
        else:
            typesCheck, optionsCheck = self.getValue(key).checks(value)
            if self.getValue(key).convertToList is True and not type(value) == list:
                value = [value]
            if all((typesCheck, optionsCheck)):
                self.getValue(key).value = value
            elif typesCheck is False:
                self.__wrong__[key] = [type(value), 'Wrong Type', self.getValue(key).types]
            elif typesCheck is True and optionsCheck is False:
                if not self.getValue(key).options == [None]:
                    self.__wrong__[key] = [type(value), 'No Option', self.getValue(key).options]
                else:
                    self.__wrong__[key] = [type(value), 'No Option', self.getValue(key).default]
        if self.__maxLen__ < len(key):
            self.__maxLen__ = len(key)

    def __getitem__(self, key: str):
        return self.getValue(key).value

    def __repr__(self) -> str:
        printString = (' ' + self.__sectionName__ + ' ').center(55,'—') + '\n'
        for i, attr in enumerate(self.list()):
            if i == self.__numAttr__-1:
                printString += "{} = {}".format(str(attr).ljust(self.__maxLen__), self.getValue(attr))
            else:
                printString += "{} = {}\n".format(str(attr).ljust(self.__maxLen__), self.getValue(attr))
        return printString


class settingsClass(parametersClass):
    def __init__(self) -> None:
        self.inputs = sectionClass('Inputs')
        self.inputs['inFile'] = 'rootSucks', 'the root input file'
        self.inputs['inPath'] = '.', 'directory of the input file'

        self.parameters = sectionClass('Parameters')
        self.parameters['eventNumbers'] = True
        self.parameters['clusterCharge'] = False
        self.parameters['seedCharge'] = False
        self.parameters['clusterSize'] = False
        self.parameters['uvSize'] = False
        self.parameters['coordinates'] = True
        self.parameters['uvPosition'] = False
        self.parameters['sensorID'] = False
        self.parameters['layer'] = False
        self.parameters['ladder'] = False
        self.parameters['pxdClusters'] = True
        self.parameters['uvCellID'] = False
        self.parameters['cellCharge'] = False
        self.parameters['clusterNumbers'] = True
        self.parameters['pdg'] = False
        self.parameters['momenta'] = False

        self.outputs = sectionClass('Outputs')
        self.outputs['outFile'] = 'rootSucks', 'the output file', str
        self.outputs['outPath'] = '.', 'directory of the output file', str
        self.outputs['saveFormat'] = 'structured', 'the output format', str, ['structured', 'array', 'txt']

        self.__allSettings__ = []
        self.__parameters__ = self.parameters.list()
        self.__userOrder__ = []
        self.__searchDict__ = {}
        for item in self.list():
            self.__allSettings__.append(self.getValue(item).list())
            for element in self.getValue(item).list():
                self.__searchDict__[element] = item
        self.__allSettings__ = flatten(self.__allSettings__)
        self.__config__ = {}

    def __getitem__(self, key):
        try:
            masterKey = self.__searchDict__[key]
            return self.getValue(masterKey)[key]
        except KeyError:
            print('{} is not an attribute of settingsClass'.format(key))

    def __setitem__(self, key, value):
        masterKey = self.__searchDict__[key]
        self.getValue(masterKey)[key] = value

    def __repr__(self) -> str:
        printString = str(self.inputs) + '\n'
        printString += str(self.parameters) + '\n'
        printString += str(self.outputs)
        return printString

    def getConfig(self, configFile=None):
        if configFile is not None:
            if os.path.isfile(configFile):
                f = open(configFile, 'r').read().split('\n')
                for line in f:
                    line = line.replace(' ', '')
                    line, _, _ = line.partition('#')
                    if len(line) == 0:
                        continue
                    line = line.replace(':', '=').split('=')
                    if line[0] in self.__parameters__:
                        self.__userOrder__.append(line[0])
                    if line[0] in self.__allSettings__:
                        if line[1] == 'true' or line[1] == 'false':
                            line[1] = line[1].capitalize()
                        self.__config__[line[0]] = getType(line[1])

                print("config file '{}' found and read".format(configFile))
            else:
                print("config file '{}' not found, continuing with defualts".format(configFile))
        else:
            print('no config specified, continuing with defualts')

    def setConfig(self):
        if 'inFile' not in self.__config__:
            pathName, _, fileName = glob('**/*.root', recursive=True)[0].partition('/')
            self.__config__['inFile'] = fileName
            self.__config__['inPath'] = pathName
        if 'outFile' not in self.__config__ and 'inFile' in self.__config__:
            self['outFile'] = self.__config__['inFile']
        if len(self.__config__) > 0:
            for line in self.__config__:
                self[line] = self.__config__[line]

    def userOrder(self):
        return self.__userOrder__
